#!/bin/sh

exec </dev/console
exec >/dev/console
exec 2>/dev/console

source /etc/env

MIN_PART0_SIZE=128
MIN_PART1_SIZE=8192
MIN_PART2_SIZE=65536
MIN_PART3_SIZE=131072
MIN_FLASH_SIZE=$(( $MIN_PART0_SIZE + $MIN_PART1_SIZE  + $MIN_PART2_SIZE + $MIN_PART3_SIZE ))

ROOTFS="rootfs.squashfs"
ROOTFS_DEBUG="rootfs.squashfs.debug"
OPTFS="optfs"
SUBFS="subfs"
DATAFS="varfs.tgz"
REMOVELIST="remove.list"
FIXENV="fixenv"
KERNEL="uImage"
RECOVERY="uImage.recovery"
FPGADATA_BB="smb_com_top.bit"
FPGADATA_RF="smb_mod.bit"
U_BOOT="u-boot.bin"
U_LOAD="u-load_01.bin"
VERSIONINFO="VersionInfo.txt"

FLASHDEV="mtdblock"
ZIPFILE="${1}"
ZIPERR="/tmp/ziperr"
SAVE="/rw/etc.tar"
RECOVERY_RSU="/recovery/recovery.rsu"

if [ "$3" -eq "$3" ] 2>/dev/null ; then
	ZIP_OFFSET="$3"
else
	ZIP_OFFSET=0
fi


setled()
{
	LED="/sys/class/leds/sgs"

	if [ "${3}" = "" ] ; then
		val="2"
	else
		val="${3}"
	fi

	echo "0" >${LED}:red:$1/brightness
	echo "0" >${LED}:green:$1/brightness

	case "${2}" in
	red)
		echo "${val}" >${LED}:red:${1}/brightness
		;;
	green)
		echo "${val}" >${LED}:green:${1}/brightness
		;;
	orange)
		echo "${val}" >${LED}:red:${1}/brightness
		echo "${val}" >${LED}:green:${1}/brightness
		;;
	off)
		;;
	esac
}

setleds()
{
	for i in $(seq 1 7)
	do
		setled ${i} ${1}
	done
}

leds_err()
{
	umount ${MOUNTPNT} 2>/dev/null
	setleds red
	setled ${1} orange
}

start_led()
{
	LEDCOLOR="${1}"
	if [ "$2" = "" ] ; then
		LEDMODE="2"
	else
		LEDMODE="${2}"
	fi
	LEDNUM="1"
	setled "${LEDNUM}" "${LEDCOLOR}" "${LEDMODE}"
}

next_led()
{
	oldled="${LEDNUM}"
	LEDNUM=$(( ${LEDNUM} + 1 ))
	if [ ${LEDNUM} -ge 8 ] ; then
		LEDNUM=1
	fi
	setled "${LEDNUM}" "${LEDCOLOR}" "${LEDMODE}"
	setled "${oldled}" "off"
}

stop_led()
{
	setled "${LEDNUM}" "off"
}

panic()
{
	echo "$1"
	umount /rw
	sgskey 60
	reboot -f
}

panic_space()
{
	[ $# -gt 1 ] && rm -rf "$2"
	leds_err 5
	panic "write error or insufficent space on storage - $1! $2"
}

file_length()
{
stat -c '%s' ${1} 2>/dev/null || echo -n -1
}
 
unzip_stdout()
{
exec 4>&1
unzip -f ${ZIP_OFFSET} -p ${ZIPFILE} "${1}" |
while true
do
	DD="$(dd bs=4194304 count=1 conv=notrunc 2>&1 >&4)"
	[ "${DD:0:14}" != "0+1 records in" ] && break
	[ "${DD:15:15}" != "0+1 records out" ] && break
	next_led
done                                                   
exec 4>&-
}

check_ziperr()
{
if [ -s "${ZIPERR}" ] ; then
	cat ${ZIPERR}
	leds_err 6
	panic "Update file corrupted!"
fi
}

zip_length()
{
next_led
echo " $(unzip -f ${ZIP_OFFSET} -l ${ZIPFILE} "${1}" 2>${ZIPERR} | grep "\<${1}$")" | tr -s ' ' | cut -f2 -d' '

check_ziperr
}

zip_read()
{
next_led
unzip_stdout "${1}" 2>${ZIPERR}
check_ziperr
}

md5sum_cmp()
{
next_led
if [ ! -e "${1}" ] ; then
	MD5_ERR="checksum file missing"
	return 1
fi
if [ $(file_length "${1}") -ne ${3} ] ; then
	MD5_ERR="wrong file length"
	return 1
fi
if [ "$(md5sum "${1}" | tr -s ' ' | cut -f1 -d' ')" != "$(zip_read ${2}.md5sum)" ] ;then
	MD5_ERR="checksum verify failed"
	return 1
fi
return 0
}

md5sum_cmp_len()
{
next_led
[ ! -e ${1} ] && return 1
test "$(dd if=${1} count=1 bs=${3} 2>/dev/null | md5sum | tr -s ' ' | cut -f1 -d' ')" = "$(zip_read ${2}.md5sum)"
return $?
}

get_part_size()
{
SIZE="$(grep '\<'${FLASHDEV}${1}'$' /proc/partitions | tr -s ' ' | cut -f4 -d' ')"

if [ "${SIZE}" = "" ] ; then
	echo 0
else
	echo "${SIZE}"
fi
}

do_repart()
{
setleds red
panic "Repartioning not available..."
}

get_mtd_erase()
{
printf "%d\n" 0x$(cat /proc/mtd | grep "^mtd0:" | cut -f3 -d' ')
}

check_md5sum()
{
	local MD5SUM

	next_led
	if [ "${2}" != "" ] ; then
		MD5SUM=$(zip_read ${1} | md5sum | tr -s ' ' | cut -f1 -d' ')

		if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
			leds_err 6
			panic "update image ${1} corrupted" 
		fi
	fi
}

check_uimage()
{
	local TMP_IMAGE=/tmp/image
	local MD5SUM

	next_led
	if [ "${3}" != "" ] ; then
		MD5SUM=$(zip_read ${1} | tee "${TMP_IMAGE}" | md5sum | tr -s ' ' | cut -f1 -d' ')

		if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
			leds_err 6
			panic "update image ${1} corrupted" 
		fi

		/sbin/ui_verify ${TMP_IMAGE} ${2} >/dev/null
		
		ret=$?

		rm -f ${TMP_IMAGE}

		return $ret
	fi

	return 1
}

update_uimage()
{
	local TMP_IMAGE=/tmp/image

	next_led
	zip_read ${1} >${TMP_IMAGE}

	UI_WRITE=$(/sbin/ui_write ${TMP_IMAGE} ${4})

	if  [ $? -ne 0 ] ; then
		leds_err 7
		panic "image update ${1} failed. ${UI_WRITE}" 
	fi

	if [ -z "${4}" ] ; then
		if ! /sbin/ui_verify ${TMP_IMAGE} =${UI_WRITE} ; then
			leds_err 7
			panic "image update ${1} verify failed."
		fi

		if ! /usr/sbin/senv ${2} ${UI_WRITE} ; then
			leds_err 7
			panic "environment update ${2} failed."
		fi
	fi

	rm -f ${TMP_IMAGE}
}


export HOME=/tmp
cd /tmp

setleds off

PART0_SIZE=$(get_part_size 0)
PART1_SIZE=$(get_part_size 1)
PART2_SIZE=$(get_part_size 2)
PART3_SIZE=$(get_part_size 3)

if [ $(( $PART0_SIZE + $PART1_SIZE  + $PART2_SIZE + $PART3_SIZE )) -lt $MIN_FLASH_SIZE ] ; then
	setleds red
	panic "insufficent flash memory"
fi

LAYOUT="$(grep '\<'${FLASHDEV}'[0-9]\+$' /proc/partitions | tr -s ' ' | cut -f5 -d' ')"
if [ "$(echo ${LAYOUT})" != "${FLASHDEV}0 ${FLASHDEV}1 ${FLASHDEV}2 ${FLASHDEV}3" ] ; then
	do_repart 0
else
	if [ $PART2_SIZE -lt $MIN_PART2_SIZE ] ; then
		do_repart 2
	else
		if [ $PART3_SIZE -lt $MIN_PART3_SIZE ] ; then
			do_repart 3
		fi
	fi
fi

start_led "orange"
if ! mount -o rw -t yaffs /dev/rw /rw ; then
	leds_err 4
	panic "cannot access data partition"
fi

# free some storage space ...
rm -rf /rw/${ROOTFS}.new /rw/${OPTFS}.new
rm -rf /rw/etc.new ${SAVE} ${SAVE}.new
if [ "$(df /rw | tail -n 1 | tr -s ' ' | cut -f4 -d' ')" -lt 10240 ] ; then
	leds_err 5
	panic "Software update failed.  At least 10 MB of free storage space is required."
fi

next_led

ROOTFS_LEN=$(zip_length ${ROOTFS})
ROOTFS_DEBUG_LEN=$(zip_length ${ROOTFS_DEBUG})

KERNEL_LEN=$(zip_length ${KERNEL})
if [ "${KERNEL_LEN}" = "" ] ; then
	panic "Cannot get update ${KERNEL}"
fi

ULOAD_LEN=$(zip_length ${U_LOAD})
FPGADATA_BB_LEN=$(zip_length ${FPGADATA_BB})
FPGADATA_RF_LEN=$(zip_length ${FPGADATA_RF})
U_BOOT_LEN=$(zip_length ${U_BOOT})
RECOVERY_LEN=$(zip_length ${RECOVERY})
OPTFS_LEN=$(zip_length ${OPTFS})
DATAFS_LEN=$(zip_length ${DATAFS})
FIXENV_LEN=$(zip_length ${FIXENV})
REMOVELIST_LEN=$(zip_length ${REMOVELIST})

echo "Checking update file..."

check_uimage ${FPGADATA_BB} fpgadata_bb_vol ${FPGADATA_BB_LEN} && FPGADATA_BB_LEN=""
check_uimage ${FPGADATA_RF} fpgadata_rf_vol ${FPGADATA_RF_LEN} && FPGADATA_RF_LEN=""
check_uimage ${U_BOOT} u_boot_vol ${U_BOOT_LEN} && U_BOOT_LEN=""
check_uimage ${RECOVERY} rescue_vol ${RECOVERY_LEN} && RECOVERY_LEN=""
if check_uimage ${KERNEL} kernel_vol ${KERNEL_LEN} ; then
	[ "${RECOVERY_LEN}" = "" ] && KERNEL_LEN=""
fi

if [ "${ULOAD_LEN}" != "" ] ; then
	check_md5sum ${U_LOAD} ${ULOAD_LEN}
	if md5sum_cmp_len /dev/mtd0 ${U_LOAD} ${ULOAD_LEN} ; then
		ULOAD_LEN=""
	fi
fi

if md5sum_cmp /rw/${ROOTFS} ${ROOTFS} ${ROOTFS_LEN} ; then
	ROOTFS_LEN=""
fi

[ "${OPTFS_LEN}" != "" ] && check_md5sum ${OPTFS} ${OPTFS_LEN}
[ "${DATAFS_LEN}" != "" ] && check_md5sum ${DATAFS} ${DATAFS_LEN}
[ "${REMOVELIST_LEN}" != "" ] && check_md5sum ${REMOVELIST} ${REMOVELIST_LEN}
[ "${FIXENV_LEN}" != "" ] && check_md5sum ${FIXENV} ${FIXENV_LEN}

next_led

mkdir /tmp/zcip
mv /rw/var/cache/zcip.* /tmp/zcip 2>/dev/null
rm -rf /rw/${SUBFS} /rw/devel /rw/testing /rw/opt /rw/startup /rw/var/cache/* 2>/dev/null
mv /tmp/zcip/zcip.* /rw/var/cache/ 2>/dev/null

grep -e "^${share_user}:" /rw/etc/passwd | head -n 1 >/rw/etc/passwd.user
grep -e "^${share_user}:" /rw/etc/samba/smbpasswd | head -n 1 >/rw/etc/samba/smbpasswd.user
tar -cf ${SAVE}.new -C /rw/etc \
 bluetooth.pin \
 rs232.conf \
 layout.conf \
 security/ \
 sysctl.conf \
 systemsecpw.srspw \
 vncpasswd \
 network/interfaces \
 samba/smbpasswd.user \
 samba/workgroup.conf \
 touch_cal \
 passwd.user \
2>/dev/null
mv ${SAVE}.new ${SAVE}
sync

next_led
pwHash=$(grep "^${share_user}:" /rw/etc/passwd | cut -d ":" -f2)
if [ -z "${pwHash}" ]; then
	pwHash=$(grep "^${share_user}:" /etc/passwd | cut -d ":" -f2)
fi
sed -e "s|^${share_user}:[^:]*:\(.*\)|${share_user}:${pwHash}:\1|" /etc/passwd > /tmp/passwd.new
mv /tmp/passwd.new /tmp/passwd

{
grep -e "^${share_user}:" -v /etc/samba/smbpasswd
{
grep -e "^${share_user}:" /rw/etc/samba/smbpasswd
grep -e "^${share_user}:" /etc/samba/smbpasswd
} | head -n 1
} >/tmp/smbpasswd

echo "Performing update..."
start_led "orange"

mv /rw/etc /rw/etc.new
rm -rf /rw/etc.new

next_led
if [ "$(penv u_boot_vol)" = "u_boot_vol=1,2,3" -a "$(penv bootenv_vol)" = "bootenv_vol=4" ] ; then
	if ! /usr/sbin/senv reserved_vol 1,2,3,4 ; then
		leds_err 7
		panic "environment update ${2} failed."
	fi
fi

next_led
if [ "${ULOAD_LEN}" != "" ] ; then
	echo "Updating boot loader..."
	flash_erase /dev/mtd0 0 1
	zip_read ${U_LOAD} |
	dd of=/dev/mtd0 bs=$(get_mtd_erase) count=1 conv=sync
	if ! md5sum_cmp_len /dev/mtd0 ${U_LOAD} ${ULOAD_LEN} ; then
		leds_err 7
		panic "Boot Loader Update failed..."
	fi
fi

[ "${U_BOOT_LEN}" != "" ] && update_uimage ${U_BOOT} u_boot_vol ${U_BOOT_LEN} u_boot_vol
[ "${FPGADATA_BB_LEN}" != "" ] && update_uimage ${FPGADATA_BB} fpgadata_bb_vol ${FPGADATA_BB_LEN}
[ "${FPGADATA_RF_LEN}" != "" ] && update_uimage ${FPGADATA_RF} fpgadata_rf_vol ${FPGADATA_RF_LEN}
[ "${RECOVERY_LEN}" != "" ] && update_uimage ${RECOVERY} rescue_vol ${RECOVERY_LEN}
[ "${KERNEL_LEN}" != "" ] && update_uimage ${KERNEL} kernel_vol ${KERNEL_LEN}

next_led
if [ "${FIXENV_LEN}" != "" ] ; then
	echo "Updating u-boot environment..."
	zip_read ${FIXENV} |
	/usr/sbin/uenv
fi

next_led
if [ "${REMOVELIST_LEN}" != "" ] ; then
	echo "Remove obsolete files..."
	zip_read ${REMOVELIST} |
	while read l
	do
		eval rm -rf ${l}
	done
fi

next_led
rm -f /rw/${OPTFS} /rw/${OPTFS}.new

next_led
if [ "${ROOTFS_LEN}" != "" ] ; then
	echo "Updating root filesystem..."
	rm -f /rw/${ROOTFS_DEBUG}
	dd if=/dev/ramdisk of=/rw/${ROOTFS}.new bs=512 count=$(( ${ROOTFS_LEN} / 512 ))
	sync
	if ! md5sum_cmp /rw/${ROOTFS}.new ${ROOTFS} ${ROOTFS_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${ROOTFS}.new
	fi
	mv -f /rw/${ROOTFS}.new /rw/${ROOTFS}
	sync
fi

next_led
if [ "${ROOTFS_DEBUG_LEN}" != "" ] ; then
	echo "Updating debug root filesystem..."
	zip_read ${ROOTFS_DEBUG} >/rw/${ROOTFS_DEBUG}.new
	sync
	if ! md5sum_cmp /rw/${ROOTFS_DEBUG}.new ${ROOTFS_DEBUG} ${ROOTFS_DEBUG_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${ROOTFS_DEBUG}.new
	fi
	mv -f /rw/${ROOTFS_DEBUG}.new /rw/${ROOTFS_DEBUG}
	sync
fi

next_led
echo "Update settings..."
(
cd /rw
mount -o loop,ro -t squashfs /rw/${ROOTFS} /ro
cp -a /ro/etc etc.new || panic_space "failed to fetch /etc" /rw/etc.new
umount /ro
losetup -d /dev/loop0 2>/dev/null
tar -xf ${SAVE} -C etc.new || panic_space "failed to restore system config" /rw/etc.new
cp /tmp/passwd etc.new/passwd || panic_space "failed to restore passwd" /rw/etc.new
cp /tmp/smbpasswd etc.new/samba/smbpasswd || panic_space "failed to restore smbpasswd" /rw/etc.new
mv etc.new etc
) >&2
sync
rm -f ${SAVE} etc/passwd.user
rm -f ${SAVE} etc/samba/smbpasswd.user

if [ "${OPTFS_LEN}" != "" ] ; then
	echo "Updating Application..."
	zip_read ${OPTFS} >/rw/${OPTFS}.new
	sync
	if ! md5sum_cmp /rw/${OPTFS}.new ${OPTFS} ${OPTFS_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${OPTFS}.new
	fi
	mv -f /rw/${OPTFS}.new /rw/${OPTFS}
	sync
fi

if [ "${DATAFS_LEN}" != "" ] ; then
	echo "Updating Application Data..."
	zip_read ${DATAFS} |
	tar -xz -C /rw
fi

next_led
sync

next_led
if mount -o rw -t yaffs /dev/recovery /recovery ; then
	if [ ! -f "${RECOVERY_RSU}" -o ! -f "${RECOVERY_RSU}.v1" ] ; then
		if [ ! -f "/recovery/${VERSIONINFO}"  -o ! -f "${RECOVERY_RSU}.v1" ] ; then
			if [ ! -f "${RECOVERY_RSU}.v1" ] ; then
				rm -rf /recovery/*
			fi
			echo "Create recovery..."
			cp ${ZIPFILE} ${RECOVERY_RSU}.save

			if cmp ${ZIPFILE} ${RECOVERY_RSU}.save ; then
				mv ${RECOVERY_RSU}.save ${RECOVERY_RSU}
			else
				rm ${RECOVERY_RSU}.save
				if mount -o loop,ro "/rw/${OPTFS}" /opt ; then
					cp -f /opt/bin/${VERSIONINFO} /recovery
					umount /opt
				fi
			fi
			touch "${RECOVERY_RSU}.v1"
		fi
	fi
	unzip_stdout ${DATAFS} >/recovery/${DATAFS} 
	umount /recovery
fi

echo "Sync Filesystems..."
next_led
sync
next_led
umount /rw

echo "Update done... rebooting system"
setleds green
sgskey 5
reboot -f

